import java.util.ArrayList;
import java.util.List;
import java.io.File;
import java.io.IOException;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.Namespace;
import org.jdom2.JDOMException;


/**
 * Class that allows for generating sensorconfigs to be used with Adapter based
 * on available gazebomodels.
 */
public class SensorConfigSetup extends Setup{
	//Variables used exclusively by SensorConfigSetup
	public final List<Document> gazeboModelList;
	public final List<String> gazeboModelNamesList;
	public final List<Document> sensorConfigList = new ArrayList<Document>();
	public final List<String> sensorConfigNamesList = new ArrayList<String>();
	
	/**
	 * Creates a runnable SensorConfigSetup.
	 * 
	 * When reading input, in_gxPath may contain other files that will be filtered out.
	 * 
	 * @throws IOException 
	 * @throws JDOMException 
	 */
	public SensorConfigSetup() throws JDOMException, IOException {
		super();
		
		//Get all gazeboModel files and file names
		gazeboModelList = new ArrayList<Document>();
		gazeboModelNamesList = new ArrayList<String>();
		
		List<String> gxList = new ArrayList<String>();
		List<String> gxDirectoryFilesList = XMLHelper.getFileNames(in_gxPath);
		for(String s : gxDirectoryFilesList) {
			if(s.endsWith(GX_ENDING)) {
				gxList.add(s);
			}
		}
		
		if(gxList.isEmpty()) {
			System.out.println("Warning: No gazebo model \'"+GX_ENDING+"\' files found in \'"+in_gxPath+
				"\'. Could not autogenerate sensorconfig files");
		}
		
		for(String gx : gxList) {
			gazeboModelList.add(XMLHelper.readXml(in_gxPath+gx));
			gazeboModelNamesList.add(gx.substring(0, gx.length() - GX_ENDING.length()));
		}
	}
	
	/**
	 * Run the SensorConfigSetup.
	 * 
	 * @throws IOException 
	 */
	public void run() throws IOException {
		System.out.println("SensorConfigSetup started.");
		
		//Remove old autogenerated sensorconfigs and create directories
		//checkDirectory(in_scPath); //Done by ImplGenerator.java
		if((new File(scAutoPath)).exists()) {
			XMLHelper.deleteDirectory(scAutoPath);
		}
		XMLHelper.checkDirectory(scAutoPath);
		XMLHelper.checkDirectory(scOtherPath);
		
		//Compute sensorconfigs for each gazeboModel
		for(int i = 0; i < gazeboModelList.size(); i++) {
			Document gazeboModel = gazeboModelList.get(i);
			String gazeboModelName = gazeboModelNamesList.get(i);
			Namespace NS = gazeboModel.getRootElement().getNamespace("xacro");
			
			//Replace all $(arg ...) variables
			this.replaceArg(gazeboModel, NS);
			
			//Generate sensorconfigs from gazeboModel
			this.generateSensorConfigs(gazeboModel, gazeboModelName);
		}	
		
		//Output sensorconfigs
		this.output();
		
		System.out.println("SensorConfigSetup finished.");
	}
	
	/**
	 * Replace all $(arg ...) variables with their concrete values.
	 * 
	 * @param gazeboModel	model to be operated on as Document
	 * @param NS			Namespace of gazeboModel as Namespace
	 */
	private void replaceArg(Document gazeboModel, Namespace NS) {
		List<Element> argList = XMLHelper.xQueryElements(gazeboModel, "//xacro:arg", NS);
		
		for(Element arg : argList) {
			String variable = "$(arg "+arg.getAttributeValue("name")+")";
			String concreteValue = arg.getAttributeValue("default");
			List<Element> occurrenceList = XMLHelper.xQueryElements(gazeboModel, "//*[text()='"+variable+"']");
			for(Element occurrence : occurrenceList) {
				occurrence.setText(concreteValue);
			}
		}
	}
	
	/**
	 * Generate a new sensorconfig for each sensor in gazeboModel.
	 * 
	 * @param gazeboModel		model to be operated on as Document
	 * @param gazeboModelName	model name as String
	 */
	private void generateSensorConfigs(Document gazeboModel, String gazeboModelName) {
		//Get all existing sensors in gazeboModel
		List<Element> sensorList = XMLHelper.xQueryElements(gazeboModel, "//sensor");
		
		//Create a new sensorconfig for each of them.
		for(Element sensor : sensorList) {	
			Element entry = sensor.getParentElement();
			if(entry.getName().equals("gazebo")) {
				entry.detach();
				
				Element sensorConfigRoot = new Element("configuration");
				Document sensorConfig = new Document(sensorConfigRoot);
				sensorConfig.getRootElement().addContent(entry);
				
				this.sensorConfigList.add(sensorConfig);
				String sensorConfigName = fixName(sensor.getAttributeValue("name"))+"___"+fixName(gazeboModelName);
				this.sensorConfigNamesList.add(sensorConfigName);
			}
			else {
				System.out.println("Found a <sensor> element that does not have a <gazebo> element as immediate parent. "
					+"It was ignored, as this input is unexpected.");	
			}
		}
	}
	
	/**
	 * Takes a String and alters it so a FeatureIDE model will accept it as a feature name.
	 * FeatureIDE says it only takes the following regex as valid feature names: ^[a-zA-Z]+\w*$
	 * This function will replace all \W characters (no word character) with '_'.
	 * 
	 * @param name		Name to be changed as String
	 * @return Corrected name
	 */
	private static String fixName(String name) {
		return name.replaceAll("\\W", "_");
	}
	
	/**
	 * Output all generated sensorconfigs as XML files.
	 * 
	 * @throws IOException 
	 */
	private void output() throws IOException {
		for(int i = 0; i < this.sensorConfigList.size(); i++) {
			Document sensorConfig = this.sensorConfigList.get(i);
			String sensorConfigName = this.sensorConfigNamesList.get(i);
			
			XMLHelper.saveAsXml(sensorConfig, this.scAutoPath+sensorConfigName+SC_ENDING);
		}
	}
}
